import { loadSchema } from '@graphql-tools/load';
import { fileURLToPath } from 'url';
import { relative } from 'pathe';
import { defu } from 'defu';
import { useLogger, resolveFiles, defineNuxtModule, resolveAlias, createResolver, addImports, addServerImports, addTypeTemplate, addTemplate, addServerHandler, addPlugin } from '@nuxt/kit';
import { onDevToolsInitialized, extendServerRpc } from '@nuxt/devtools-kit';
import { existsSync } from 'fs';
import { GraphqlMiddlewareTemplate } from '../dist/runtime/settings/index.js';
import fs from 'node:fs/promises';
import { existsSync as existsSync$1, promises } from 'node:fs';
import { generate } from '@graphql-codegen/cli';
import * as PluginSchemaAst from '@graphql-codegen/schema-ast';
import { basename } from 'node:path';
import { printSourceLocation, parse, OperationTypeNode, Source } from 'graphql';
import { Generator, FieldNotFoundError, TypeNotFoundError, FragmentNotFoundError } from 'graphql-typescript-deluxe';
import colors from 'picocolors';
import { validateGraphQlDocuments } from '@graphql-tools/utils';

const name = "nuxt-graphql-middleware";
const version = "5.0.0-alpha.5";

const DEVTOOLS_UI_ROUTE = "/__nuxt-graphql-middleware";
const DEVTOOLS_UI_LOCAL_PORT = 3300;
function setupDevToolsUI(nuxt, clientPath) {
  const isProductionBuild = existsSync(clientPath);
  if (isProductionBuild) {
    nuxt.hook("vite:serverCreated", async (server) => {
      const sirv = await import('sirv').then((r) => r.default || r);
      server.middlewares.use(
        DEVTOOLS_UI_ROUTE,
        sirv(clientPath, { dev: true, single: true })
      );
    });
  } else {
    nuxt.hook("vite:extendConfig", (config) => {
      config.server = config.server || {};
      config.server.proxy = config.server.proxy || {};
      config.server.proxy[DEVTOOLS_UI_ROUTE] = {
        target: "http://localhost:" + DEVTOOLS_UI_LOCAL_PORT + DEVTOOLS_UI_ROUTE,
        changeOrigin: true,
        followRedirects: true,
        rewrite: (path) => path.replace(DEVTOOLS_UI_ROUTE, "")
      };
    });
  }
  nuxt.hook("devtools:customTabs", (tabs) => {
    tabs.push({
      // unique identifier
      name: "nuxt-graphql-middleware",
      // title to display in the tab
      title: "GraphQL Middleware",
      // any icon from Iconify, or a URL to an image
      icon: "akar-icons:graphql-fill",
      // iframe view
      view: {
        type: "iframe",
        src: DEVTOOLS_UI_ROUTE
      }
    });
  });
}

function pluginLoader(name) {
  switch (name) {
    case "@graphql-codegen/schema-ast":
      return Promise.resolve(PluginSchemaAst);
  }
  throw new Error(`graphql-codegen plugin not found: ${name}`);
}
function generateSchema(moduleOptions, dest, writeToDisk) {
  const pluginConfig = moduleOptions.codegenSchemaConfig?.urlSchemaOptions;
  const schemaAstConfig = moduleOptions.codegenSchemaConfig?.schemaAstConfig || {
    sort: true
  };
  const config = {
    schema: moduleOptions.graphqlEndpoint,
    pluginLoader,
    silent: true,
    errorsOnly: true,
    config: pluginConfig,
    generates: {
      [dest]: {
        plugins: ["schema-ast"],
        config: schemaAstConfig
      }
    }
  };
  return generate(config, writeToDisk).then((v) => v[0]);
}

const logger = useLogger(name);
const defaultOptions = {
  downloadSchema: true,
  schemaPath: "~~/schema.graphql",
  serverApiPrefix: "/api/graphql_middleware",
  graphqlEndpoint: "",
  debug: false,
  includeComposables: true,
  documents: [],
  devtools: true,
  errorOverlay: true
};
function validateOptions(options) {
  if (!options.graphqlEndpoint) {
    throw new Error("Missing graphqlEndpoint.");
  }
}
async function getSchemaPath(schemaPath, options, resolver, writeToDisk = false) {
  const dest = resolver(schemaPath);
  if (!options.downloadSchema) {
    const fileExists2 = await fs.access(dest).then(() => true).catch(() => false);
    if (!fileExists2) {
      logger.error(
        '"downloadSchema" is set to false but no schema exists at ' + dest
      );
      throw new Error("Missing GraphQL schema.");
    }
    const schemaContent = await fs.readFile(dest).then((v) => v.toString());
    return { schemaPath, schemaContent };
  }
  if (!options.graphqlEndpoint) {
    throw new Error("Missing graphqlEndpoint config.");
  }
  const result = await generateSchema(options, dest, writeToDisk);
  return { schemaPath, schemaContent: result.content };
}
const fileExists = (path, extensions = ["js", "ts", "mjs"]) => {
  if (!path) {
    return null;
  } else if (existsSync$1(path)) {
    return path;
  }
  const extension = extensions.find(
    (extension2) => existsSync$1(`${path}.${extension2}`)
  );
  return extension ? `${path}.${extension}` : null;
};

function generateResponseTypeTemplate(operations, context) {
  const allTypes = operations.map((v) => v.typeName).sort();
  return `import type {
  ${allTypes.join(",\n  ")}
} from './../graphql-operations'
import type { GraphqlResponseAdditions } from './server-options'
import type { GraphqlServerResponse } from '${context.runtimeTypesPath}'

export type GraphqlMiddlewareResponseUnion =
  | ${allTypes.join("\n  | ") || "never"}

export type GraphqlResponse<T> = GraphqlServerResponse<T> & GraphqlResponseAdditions
export type GraphqlResponseTyped = GraphqlResponse<GraphqlMiddlewareResponseUnion>`;
}

const SYMBOL_CROSS = "x";
const SYMBOL_CHECK = "\u2714";
function getMaxLengths(entries) {
  let name = 0;
  let path = 0;
  let type = 0;
  for (const entry of entries) {
    if (entry.type.length > type) {
      type = entry.type.length;
    }
    if (entry.name.length > name) {
      name = entry.name.length;
    }
    if (entry.path.length > path) {
      path = entry.path.length;
    }
  }
  return { name, path, type };
}
function logAllEntries(entries) {
  const lengths = getMaxLengths(entries);
  let prevHadError = false;
  for (const entry of entries) {
    const hasErrors = entry.errors.length > 0;
    const icon = hasErrors ? colors.red(SYMBOL_CROSS) : colors.green(SYMBOL_CHECK);
    const type = entry.type.padEnd(lengths.type);
    const namePadded = colors.bold(entry.name.padEnd(lengths.name));
    const name = hasErrors ? colors.red(namePadded) : colors.green(namePadded);
    const path = colors.dim(entry.path);
    const parts = [icon, type, name, path];
    if (hasErrors && !prevHadError) {
      process.stdout.write("-".repeat(process.stdout.columns) + "\n");
    }
    logger.log(parts.join(" | "));
    if (hasErrors) {
      const errorLines = [];
      entry.errors.forEach((error) => {
        let output = colors.red(error.message);
        if (error.source && error.locations) {
          for (const location of error.locations) {
            output += "\n\n" + colors.red(printSourceLocation(error.source, location));
          }
        }
        errorLines.push(output);
      });
      logger.log(
        errorLines.join("\n").split("\n").map((v) => "    " + v).join("\n")
      );
      process.stdout.write("-".repeat(process.stdout.columns) + "\n");
    }
    prevHadError = hasErrors;
  }
  logger.restoreStd();
}

class CollectedFile {
  filePath;
  fileContents;
  isOnDisk;
  parsed;
  constructor(filePath, fileContents, isOnDisk = false) {
    this.filePath = filePath;
    this.fileContents = fileContents;
    this.isOnDisk = isOnDisk;
    this.parsed = parse(fileContents);
  }
  static async fromFilePath(filePath) {
    const content = (await promises.readFile(filePath)).toString();
    return new CollectedFile(filePath, content, true);
  }
  /**
   * If isOnDisk, re-read file contents from disk, then parse it (syntax only).
   */
  async update() {
    if (this.isOnDisk) {
      const newContents = (await promises.readFile(this.filePath)).toString();
      if (newContents === this.fileContents) {
        return false;
      }
      this.fileContents = newContents;
      this.parsed = parse(newContents);
      return true;
    }
    return false;
  }
}

function generateNitroTypes(operations, serverApiPrefix) {
  const endpoints = [];
  const imports = [];
  for (const operation of operations) {
    imports.push(operation.typeName);
    const method = operation.operationType === OperationTypeNode.QUERY ? "get" : "post";
    endpoints.push(
      `    '${serverApiPrefix}/${operation.operationType}/${operation.graphqlName}': {
      '${method}': GraphqlResponse<${operation.typeName}>
    }`
    );
  }
  return `import type { GraphqlResponse } from './response'
import type {
  ${imports.sort().join(",\n  ")}
} from './../graphql-operations'

declare module 'nitropack/types' {
  interface InternalApi {
${endpoints.sort().join("\n")}
  }
}`;
}

function generateSourcesTemplate(operations, srcDir) {
  const lines = [];
  for (const operation of operations) {
    const filePath = relative(srcDir, operation.filePath);
    lines.push(
      `${operation.operationType}_${operation.graphqlName}: '${filePath}',`
    );
  }
  return `
export const operationSources = {
  ${lines.join("\n  ")}
}
`;
}

class Collector {
  constructor(schema, context, nuxtConfigDocuments = [], generatorOptions = {}) {
    this.schema = schema;
    this.context = context;
    this.nuxtConfigDocuments = nuxtConfigDocuments;
    const mappedOptions = { ...generatorOptions };
    if (!mappedOptions.output) {
      mappedOptions.output = {};
    }
    if (!mappedOptions.output.buildTypeDocFilePath) {
      mappedOptions.output.buildTypeDocFilePath = (filePath) => {
        return this.filePathToBuildRelative(filePath);
      };
    }
    this.generator = new Generator(schema, mappedOptions);
  }
  /**
   * All collected files.
   */
  files = /* @__PURE__ */ new Map();
  /**
   * The code generator.
   */
  generator;
  /**
   * A map of operation name and timestamp when the operation was last validated.
   */
  operationTimestamps = /* @__PURE__ */ new Map();
  /**
   * The generated operations and fragments.
   */
  rpcItems = /* @__PURE__ */ new Map();
  /**
   * The generated TypeScript type template output.
   */
  outputTypes = "";
  /**
   * The generated TypeScript enum template output.
   */
  outputEnums = "";
  /**
   * The generated oeprations file.
   */
  outputOperations = "";
  /**
   * The generated oepration types file.
   */
  outputOperationTypes = "";
  /**
   * The generated context template file.
   */
  outputResponseTypes = "";
  /**
   * The generated nitro template file.
   */
  outputNitroTypes = "";
  /**
   * The generated nitro template file.
   */
  outputSources = "";
  filePathToBuildRelative(filePath) {
    return "./" + relative(this.context.buildDir, filePath);
  }
  filePathToSourceRelative(filePath) {
    return "./" + relative(process.cwd(), filePath);
  }
  operationToLogEntry(operation, errors) {
    return {
      name: operation.graphqlName,
      type: operation.operationType,
      path: this.filePathToSourceRelative(operation.filePath),
      errors
    };
  }
  buildOutputTypes(file) {
    let output = "";
    const enumImports = file.getTypeScriptEnumDependencies();
    if (enumImports.length) {
      output += `import type { ${enumImports.join(", ")} } from './enums'

`;
    }
    output += file.getSource();
    return output;
  }
  /**
   * Executes code gen and performs validation for operations.
   */
  buildState() {
    const output = this.generator.build();
    const operations = output.getCollectedOperations();
    const generatedCode = output.getGeneratedCode();
    this.outputOperations = output.getOperationsFile({
      exportName: "documents",
      minify: !this.context.isDev
    }).getSource();
    this.outputOperationTypes = output.getOperationTypesFile({
      importFrom: "./../graphql-operations"
    }).getSource();
    this.outputEnums = output.buildFile(["enum"]).getSource();
    this.outputTypes = this.buildOutputTypes(output.getTypes());
    this.outputResponseTypes = generateResponseTypeTemplate(
      operations,
      this.context
    );
    this.outputNitroTypes = generateNitroTypes(
      operations,
      this.context.serverApiPrefix
    );
    this.outputSources = generateSourcesTemplate(
      operations,
      this.context.rootDir
    );
    const fragmentMap = /* @__PURE__ */ new Map();
    const operationSourceMap = /* @__PURE__ */ new Map();
    for (const code of generatedCode) {
      if (code.type === "fragment" && code.graphqlName && code.source) {
        fragmentMap.set(code.graphqlName, code.source);
      } else if (code.type === "operation" && code.graphqlName && code.source) {
        operationSourceMap.set(code.graphqlName, code.source);
      }
    }
    let hasErrors = false;
    const logEntries = [];
    for (const operation of operations) {
      const previousTimestamp = this.operationTimestamps.get(
        operation.graphqlName
      );
      if (previousTimestamp === operation.timestamp) {
        continue;
      }
      const fragments = operation.getGraphQLFragmentDependencies().map((v) => fragmentMap.get(v) || "").join("\n");
      const fullOperation = operationSourceMap.get(operation.graphqlName) + fragments;
      const source = new Source(fullOperation, basename(operation.filePath));
      const document = parse(source);
      const errors = validateGraphQlDocuments(this.schema, [document]);
      if (errors.length) {
        hasErrors = true;
      } else {
        this.operationTimestamps.set(operation.graphqlName, operation.timestamp);
      }
      const shouldLog = errors.length || !this.context.logOnlyErrors;
      if (shouldLog) {
        logEntries.push(this.operationToLogEntry(operation, errors));
      }
    }
    logAllEntries(logEntries);
    if (hasErrors) {
      throw new Error("GraphQL errors");
    }
    for (const code of generatedCode) {
      const id = `${code.identifier}_${code.graphqlName}`;
      if (code.identifier === "fragment" || code.identifier === "mutation" || code.identifier === "query") {
        if (this.rpcItems.get(id)?.timestamp === code.timestamp) {
          continue;
        }
        const fragmentDepdendencies = code.getGraphQLFragmentDependencies().map((name) => fragmentMap.get(name) || "").join("\n\n");
        this.rpcItems.set(id, {
          id,
          timestamp: code.timestamp,
          source: code.source + "\n\n" + fragmentDepdendencies,
          name: code.graphqlName,
          filePath: code.filePath,
          identifier: code.identifier
        });
      }
    }
  }
  buildErrorMessage(error) {
    let output = "";
    if (error instanceof FieldNotFoundError || error instanceof TypeNotFoundError || error instanceof FragmentNotFoundError) {
      const filePath = error.context?.filePath;
      const file = filePath ? this.files.get(filePath) : null;
      if (filePath) {
        output += ` | ${this.filePathToSourceRelative(filePath)}
`;
      }
      output += "\n" + error.message + "\n\n";
      if (file) {
        output += file.fileContents;
      }
    } else if (error instanceof Error) {
      output += "\n" + error.message;
    }
    return output;
  }
  logError(error) {
    let output = `${SYMBOL_CROSS}`;
    output += this.buildErrorMessage(error);
    logger.error(colors.red(output));
  }
  /**
   * Get all file paths that match the import patterns.
   */
  async getImportPatternFiles() {
    if (this.context.patterns.length) {
      return resolveFiles(this.context.srcDir, this.context.patterns, {
        followSymbolicLinks: false
      });
    }
    return [];
  }
  /**
   * Initialise the collector.
   */
  async init() {
    try {
      const files = await this.getImportPatternFiles();
      for (const filePath of files) {
        await this.addFile(filePath);
      }
      const nuxtConfigDocuments = this.nuxtConfigDocuments.join("\n\n");
      if (nuxtConfigDocuments.length) {
        const filePath = this.context.nuxtConfigPath;
        const file = new CollectedFile(filePath, nuxtConfigDocuments, false);
        this.files.set(filePath, file);
        this.generator.add({
          filePath,
          documentNode: file.parsed
        });
      }
      this.buildState();
      logger.success("All GraphQL documents are valid.");
    } catch (e) {
      this.logError(e);
      throw new Error("GraphQL document validation failed.");
    }
  }
  /**
   * Add a file.
   */
  async addFile(filePath) {
    const file = await CollectedFile.fromFilePath(filePath);
    this.files.set(filePath, file);
    this.generator.add({
      filePath,
      documentNode: file.parsed
    });
    return file;
  }
  async handleAdd(filePath) {
    const matching = await this.getImportPatternFiles();
    if (!matching.includes(filePath)) {
      return false;
    }
    await this.addFile(filePath);
    return true;
  }
  async handleChange(filePath) {
    const file = this.files.get(filePath);
    if (!file) {
      return false;
    }
    const needsUpdate = await file.update();
    if (!needsUpdate) {
      return false;
    }
    this.generator.update({
      filePath,
      documentNode: file.parsed
    });
    return true;
  }
  handleUnlink(filePath) {
    const file = this.files.get(filePath);
    if (!file) {
      return false;
    }
    this.files.delete(filePath);
    this.generator.remove(filePath);
    return true;
  }
  handleUnlinkDir(folderPath) {
    let anyHasChanged = false;
    for (const filePath of [...this.files.keys()]) {
      if (filePath.startsWith(folderPath)) {
        const hasChanged = this.handleUnlink(filePath);
        if (hasChanged) {
          anyHasChanged = true;
        }
      }
    }
    return anyHasChanged;
  }
  /**
   * Handle the watcher event for the given file path.
   */
  async handleWatchEvent(event, filePath) {
    let hasChanged = false;
    try {
      if (event === "add") {
        hasChanged = await this.handleAdd(filePath);
      } else if (event === "change") {
        hasChanged = await this.handleChange(filePath);
      } else if (event === "unlink") {
        hasChanged = this.handleUnlink(filePath);
      } else if (event === "unlinkDir") {
        hasChanged = this.handleUnlinkDir(filePath);
      } else if (event === "addDir") {
      }
      if (hasChanged) {
        this.buildState();
      }
    } catch (e) {
      this.generator.resetCaches();
      logger.error("Failed to update GraphQL code.");
      this.logError(e);
      return {
        hasChanged: false,
        error: { message: this.buildErrorMessage(e) }
      };
    }
    if (hasChanged) {
      logger.success("Finished GraphQL code update successfully.");
    }
    return { hasChanged };
  }
  /**
   * Get the TypeScript types template contents.
   */
  getTemplateTypes() {
    return this.outputTypes;
  }
  /**
   * Get the TypeScript Enums template contents.
   */
  getTemplateEnums() {
    return this.outputEnums;
  }
  /**
   * Get the context template contents.
   */
  getTemplateResponseTypes() {
    return this.outputResponseTypes;
  }
  /**
   * Get the operations template contents.
   */
  getTemplateOperations() {
    return this.outputOperations;
  }
  /**
   * Get the operation types template contents.
   */
  getTemplateOperationTypes() {
    return this.outputOperationTypes;
  }
  /**
   * Get the nitro types template contents.
   */
  getTemplateNitroTypes() {
    return this.outputNitroTypes;
  }
  /**
   * Get the nitro types template contents.
   */
  getTemplateSources() {
    return this.outputSources;
  }
}

function generateDocumentTypesTemplate() {
  return `
import type { Query, Mutation } from './operations'

declare module '#nuxt-graphql-middleware/documents' {
  export type Documents = {
    query: Record<keyof Query, string>
    mutation: Record<keyof Mutation, string>
  }
  export const documents: Documents
}`;
}

function useViteWebSocket(nuxt) {
  return new Promise((resolve) => {
    nuxt.hooks.hook("vite:serverCreated", (viteServer) => {
      resolve(viteServer.ws);
    });
  });
}
const RPC_NAMESPACE = "nuxt-graphql-middleware";
const module = defineNuxtModule({
  meta: {
    name,
    configKey: "graphqlMiddleware",
    version,
    compatibility: {
      nuxt: ">=3.15.0"
    }
  },
  defaults: defaultOptions,
  async setup(passedOptions, nuxt) {
    const options = defu({}, passedOptions, defaultOptions);
    function addAlias(name2, path) {
      nuxt.options.alias[name2] = path;
    }
    function inlineNitroExternals(path) {
      nuxt.options.nitro.externals = nuxt.options.nitro.externals || {};
      nuxt.options.nitro.externals.inline = nuxt.options.nitro.externals.inline || [];
      nuxt.options.nitro.externals.inline.push(path);
    }
    const isModuleBuild = process.env.MODULE_BUILD === "true" && nuxt.options._prepare;
    if (isModuleBuild) {
      options.graphqlEndpoint = "http://localhost";
      options.downloadSchema = false;
      options.schemaPath = "~~/schema.graphql";
      options.autoImportPatterns = [
        "~~/playground/**/*.{gql,graphql}",
        "!node_modules"
      ];
    }
    if (!passedOptions.autoImportPatterns) {
      options.autoImportPatterns = ["~~/**/*.{gql,graphql}", "!node_modules"];
    }
    options.autoImportPatterns = (options.autoImportPatterns || []).map(
      (pattern) => {
        return resolveAlias(pattern);
      }
    );
    if (!nuxt.options._prepare) {
      validateOptions(options);
    }
    const moduleResolver = createResolver(import.meta.url);
    const serverResolver = createResolver(nuxt.options.serverDir);
    const srcResolver = createResolver(nuxt.options.srcDir);
    const appResolver = createResolver(nuxt.options.dir.app);
    const rootDir = nuxt.options.rootDir;
    const rootResolver = createResolver(rootDir);
    const { schemaPath, schemaContent } = await getSchemaPath(
      resolveAlias(options.schemaPath),
      options,
      rootResolver.resolve,
      options.downloadSchema
    );
    const schema = await loadSchema(schemaContent, {
      loaders: []
    });
    const runtimeDir = fileURLToPath(new URL("./runtime", import.meta.url));
    nuxt.options.build.transpile.push(runtimeDir);
    const nuxtGraphqlMiddlewareBuildDir = nuxt.options.buildDir + "/nuxt-graphql-middleware";
    const operationTypesBuildDir = nuxt.options.buildDir + "/graphql-operations";
    const toBuildRelative = (path) => {
      return relative(nuxtGraphqlMiddlewareBuildDir, path);
    };
    addAlias("#nuxt-graphql-middleware", nuxtGraphqlMiddlewareBuildDir);
    addAlias("#graphql-operations", operationTypesBuildDir);
    const context = {
      isDev: nuxt.options.dev,
      patterns: options.autoImportPatterns || [],
      srcDir: nuxt.options.srcDir,
      rootDir: nuxt.options.rootDir,
      buildDir: srcResolver.resolve(nuxt.options.buildDir),
      nuxtConfigPath: rootResolver.resolve("nuxt.config.ts"),
      schemaPath,
      serverApiPrefix: options.serverApiPrefix,
      logOnlyErrors: !!options.logOnlyErrors,
      runtimeTypesPath: toBuildRelative(
        moduleResolver.resolve("./runtime/types.ts")
      )
    };
    const collector = new Collector(
      schema,
      context,
      options.documents,
      options.codegenConfig
    );
    await collector.init();
    const isDevToolsEnabled = nuxt.options.dev && options.devtools;
    let rpc;
    if (isDevToolsEnabled) {
      const clientPath = moduleResolver.resolve("./client");
      setupDevToolsUI(nuxt, clientPath);
      onDevToolsInitialized(() => {
        rpc = extendServerRpc(RPC_NAMESPACE, {
          // register server RPC functions
          getModuleOptions() {
            return options;
          },
          getDocuments() {
            return [...collector.rpcItems.values()];
          }
        });
      });
    }
    nuxt.options.appConfig.graphqlMiddleware = {
      clientCacheEnabled: !!options.clientCache?.enabled,
      clientCacheMaxSize: options.clientCache?.maxSize || 100
    };
    nuxt.options.runtimeConfig.graphqlMiddleware = {
      graphqlEndpoint: options.graphqlEndpoint || ""
    };
    if (options.includeComposables) {
      const nuxtComposables = [
        "useGraphqlQuery",
        "useGraphqlMutation",
        "useGraphqlState",
        "useAsyncGraphqlQuery"
      ];
      if (options.enableFileUploads) {
        nuxtComposables.push("useGraphqlUploadMutation");
      }
      nuxtComposables.forEach((name2) => {
        addImports({
          from: moduleResolver.resolve("./runtime/composables/" + name2),
          name: name2
        });
      });
      const serverUtils = ["useGraphqlQuery", "useGraphqlMutation"].map(
        (name2) => {
          return {
            from: moduleResolver.resolve("./runtime/server/utils/" + name2),
            name: name2
          };
        }
      );
      addServerImports(serverUtils);
    }
    addTypeTemplate({
      filename: GraphqlMiddlewareTemplate.OperationTypes,
      write: true,
      getContents: () => collector.getTemplateTypes()
    });
    addTypeTemplate({
      filename: GraphqlMiddlewareTemplate.Types,
      write: true,
      getContents: () => {
        return `
declare module '#nuxt-graphql-middleware/sources' {
  export const operationSources: Record<string, string>
}
`;
      }
    });
    addTemplate({
      filename: GraphqlMiddlewareTemplate.Enums,
      write: true,
      getContents: () => collector.getTemplateEnums()
    });
    addTemplate({
      filename: GraphqlMiddlewareTemplate.OperationSources,
      write: true,
      getContents: () => collector.getTemplateSources()
    });
    addTemplate({
      filename: GraphqlMiddlewareTemplate.Helpers,
      write: true,
      getContents: () => `export const serverApiPrefix = '${context.serverApiPrefix}'
export function getEndpoint(operation, operationName) {
  return '${context.serverApiPrefix}' + '/' + operation + '/' + operationName
}
`
    });
    addTypeTemplate({
      filename: GraphqlMiddlewareTemplate.HelpersTypes,
      write: true,
      getContents: () => `export const serverApiPrefix: string;
export function getEndpoint(operation: string, operationName: string): string
`
    });
    addTypeTemplate({
      filename: GraphqlMiddlewareTemplate.NitroTypes,
      write: true,
      getContents: () => collector.getTemplateNitroTypes()
    });
    addTypeTemplate({
      filename: GraphqlMiddlewareTemplate.OperationTypesAll,
      write: true,
      getContents: () => collector.getTemplateOperationTypes()
    });
    const templateDocuments = addTemplate({
      filename: GraphqlMiddlewareTemplate.Documents,
      write: true,
      getContents: () => collector.getTemplateOperations()
    });
    inlineNitroExternals(templateDocuments.dst);
    addTypeTemplate({
      filename: GraphqlMiddlewareTemplate.ResponseTypes,
      write: true,
      getContents: () => collector.getTemplateResponseTypes()
    });
    addTypeTemplate({
      write: true,
      filename: "nuxt-graphql-middleware/documents.d.ts",
      getContents: () => generateDocumentTypesTemplate()
    });
    const findServerOptions = () => {
      const newPath = serverResolver.resolve("graphqlMiddleware.serverOptions");
      const serverPath = fileExists(newPath);
      if (serverPath) {
        return serverPath;
      }
      const candidates = [
        rootResolver.resolve("graphqlMiddleware.serverOptions"),
        rootResolver.resolve("app/graphqlMiddleware.serverOptions"),
        srcResolver.resolve("graphqlMiddleware.serverOptions")
      ];
      for (let i = 0; i < candidates.length; i++) {
        const path = candidates[i];
        const filePath = fileExists(path);
        if (filePath) {
          logger.warn(
            `The graphqlMiddleware.serverOptions file should be placed in Nuxt's <serverDir> ("${nuxt.options.serverDir}/graphqlMiddleware.serverOptions.ts"). The new path will be enforced in the next major release.`
          );
          return filePath;
        }
      }
      logger.info("No graphqlMiddleware.serverOptions file found.");
    };
    const resolvedServerOptionsPath = findServerOptions();
    const moduleTypesPath = toBuildRelative(moduleResolver.resolve("./types"));
    const resolvedPathRelative = resolvedServerOptionsPath ? toBuildRelative(resolvedServerOptionsPath) : null;
    const template = addTemplate({
      filename: "nuxt-graphql-middleware/server-options.mjs",
      write: true,
      getContents: () => {
        const serverOptionsLine = resolvedPathRelative ? `import serverOptions from '${resolvedPathRelative}'` : `const serverOptions = {}`;
        return `
${serverOptionsLine}
export { serverOptions }
`;
      }
    });
    inlineNitroExternals(template.dst);
    addTemplate({
      filename: "nuxt-graphql-middleware/server-options.d.ts",
      write: true,
      getContents: () => {
        const serverOptionsLineTypes = resolvedPathRelative ? `import serverOptions from '${resolvedPathRelative}'` : `const serverOptions: GraphqlMiddlewareServerOptions = {}`;
        return `
import type { GraphqlMiddlewareServerOptions } from '${moduleTypesPath}'
${serverOptionsLineTypes}

export type GraphqlResponseAdditions =
  typeof serverOptions extends GraphqlMiddlewareServerOptions<infer R, any, any> ? R : {}

export { serverOptions }`;
      }
    });
    const getClientOptionsImport = () => {
      const clientOptionsPath = appResolver.resolve(
        "graphqlMiddleware.clientOptions"
      );
      if (fileExists(clientOptionsPath)) {
        const pathRelative = toBuildRelative(clientOptionsPath);
        return `import clientOptions from '${pathRelative}'`;
      }
    };
    const clientOptionsImport = getClientOptionsImport();
    addTemplate({
      filename: "nuxt-graphql-middleware/client-options.mjs",
      write: true,
      getContents: () => {
        if (clientOptionsImport) {
          return `${clientOptionsImport}
export { clientOptions }`;
        }
        return `export const clientOptions = {}`;
      }
    });
    addTemplate({
      filename: "nuxt-graphql-middleware/client-options.d.ts",
      write: true,
      getContents: () => {
        if (clientOptionsImport) {
          return `import type { GraphqlClientOptions } from '${context.runtimeTypesPath}'
${clientOptionsImport}

export type GraphqlClientContext = typeof clientOptions extends GraphqlClientOptions<infer R> ? R : {}

export { clientOptions }`;
        }
        return `import type { GraphqlClientOptions } from '${context.runtimeTypesPath}'
export const clientOptions: GraphqlClientOptions

export type GraphqlClientContext = {}
`;
      }
    });
    addServerHandler({
      handler: moduleResolver.resolve("./runtime/serverHandler/index"),
      route: options.serverApiPrefix + "/:operation/:name"
    });
    if (options.enableFileUploads) {
      addServerHandler({
        handler: moduleResolver.resolve("./runtime/serverHandler/upload"),
        route: options.serverApiPrefix + "/upload/:name"
      });
    }
    addPlugin(moduleResolver.resolve("./runtime/plugins/provideState"), {
      append: false
    });
    if (context.isDev && options.errorOverlay) {
      addPlugin(moduleResolver.resolve("./runtime/plugins/devMode"), {
        append: false
      });
    }
    nuxt.hook("nitro:config", (nitroConfig) => {
      nitroConfig.externals = defu(
        typeof nitroConfig.externals === "object" ? nitroConfig.externals : {},
        {
          inline: [moduleResolver.resolve("./runtime")]
        }
      );
    });
    if (nuxt.options.dev || nuxt.options._prepare) {
      let sendError = function(error) {
        wsPromise.then((ws) => {
          ws.send({
            type: "error",
            err: {
              message: error.message,
              stack: ""
            }
          });
        });
      };
      addServerHandler({
        handler: moduleResolver.resolve("./runtime/serverHandler/debug"),
        route: options.serverApiPrefix + "/debug"
      });
      const wsPromise = useViteWebSocket(nuxt);
      nuxt.hook("builder:watch", async (event, pathAbsolute) => {
        if (pathAbsolute === schemaPath) {
          return;
        }
        if (!pathAbsolute.match(/\.(gql|graphql)$/)) {
          return;
        }
        const { hasChanged, error } = await collector.handleWatchEvent(
          event,
          pathAbsolute
        );
        if (error) {
          sendError(error);
        }
        if (hasChanged) {
          if (rpc) {
            rpc.broadcast.documentsUpdated([...collector.rpcItems.values()]);
          }
        }
      });
    }
  }
});

export { module as default };
